#include <Wire.h> // Arduino Wire library for I2C
#include <LiquidCrystal_I2C.h> // LCD display library for I2C
#include <Keypad.h> // Keypad library
#include <LedControl.h> // LED matrix library
#include <Servo.h>

Servo myServo;  // create a servo object
// Constants for row and column sizes
const byte ROWS = 3; 
const byte COLS = 3; 


// Array to represent keys on keypad
char hexaKeys[ROWS][COLS] = {
  {'1', '2', '3'},
  {'4', '5', '6'},
  {'7', '8', '9'}
};

// Connections to Arduino
byte rowPins[ROWS] = {7, 6, 5};
byte colPins[COLS] = {4, 3, 2};

Keypad customKeypad = Keypad(makeKeymap(hexaKeys), rowPins, colPins, ROWS, COLS);

LiquidCrystal_I2C lcd(0x27 , 16, 2); // 0x27 is the I2C address, 16 rows and 2 columns

LedControl lc = LedControl(12, 11, 10, 1); // we use pins 12, 11, 10 for the SPI interface : pin 12 -> DATA IN-pin, pin 11 -> CLK-pin, pin 10 -> CS
                                          //* there will only be a single MAX7219 attached to the arduino 
int xPos, yPos;
int PLAYER;
int WON;
const int buzzer = 9; //buzzer to arduino pin 9
int scoreP0 = 0;
int scoreP1 = 0;

//  maps for player A and player B
int A[8][8] = {
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0}
};
int B[8][8] = {
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0}
};

// hit maps
int hitA[8][8] = {
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0}
};
int hitB[8][8] = {
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0}
};

// miss maps
int missA[8][8] = {
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0}
};
int missB[8][8] = {
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0},
  {0, 0, 0, 0, 0, 0, 0, 0}
};

// BACKUP MAPS
int saveA[8][8];
int saveB[8][8];

void setup() {
  const int buzzer = 9; //buzzer to arduino pin 9
  randomSeed(analogRead(0)); // seeds the random number generator by a value obtained from doing AnalogRead on pin A0
  myServo.attach(13);   // attaches the servo on pin 9 to the servo object
  lcd.backlight(); // turn on backlight
  //lcd.init();    
  lcd.begin(); // initialize the lcd

  lc.shutdown(0, false);   // turns the display on 
  lc.setIntensity(0, 10); // set the brightness
  lc.clearDisplay(0);    // clear the display

  Serial.begin(9600);  // initialize serial communication

  // start position for attack
  xPos = 0;
  yPos = 7;

  // display on LED matrix
  lc.setRow(0, 0, B10000000); //setting the leds from the first row (index=0) 

  // first player - player 0
  PLAYER = 0;
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("PLAYER #0");
  displayScore();
  delay(700);

  // generate maps and backup maps
  generateMaps();

  // game is won when WON != -1
  WON = -1;
}

void cheat(){
  
    for (int i = 0; i < 8; i++) {
      for (int j = 0; j < 8; j++) {
        A[i][j]=0;
        B[i][j]=1;
        }
      }

      A[7][0]=1;
    

}

void winSound(){

    tone(9,988,100);
  delay(100);
  tone(9,1319,350);
  delay(500);
  noTone(9);
  delay(200);
    tone(9,988,100);
  delay(100);
  tone(9,1319,350);
  delay(500);
  noTone(9);
  delay(200);
    tone(9,988,100);
  delay(100);
  tone(9,1319,350);
  delay(500);
  noTone(9);
  delay(1000);
  
}

// functions used to check if 2 segments intersect
// source: https://www.geeksforgeeks.org/check-if-two-given-line-segments-intersect/

int getMax (int x, int y) {
  if (x > y) return x;
  return y;
}
int getMin (int x, int y) {
  if (x < y)  return x;
  return y;
}

// Given three colinear points p, q, r, the function checks if
// point q lies on line segment 'pr'
int onSegment(int px, int py, int qx, int qy, int rx, int ry)
{
  if (qx <= getMax(px, rx) && qx >= getMin(px, rx) &&
      qy <= getMax(py, ry) && qy >= getMin(py, ry))
    return 1;  // true - point q lies on line segment 'pr'

  return 0; // false
}
// To find orientation of ordered triplet (p, q, r).
// The function returns following values
// 0 --> p, q and r are colinear
// 1 --> Clockwise
// 2 --> Counterclockwise
int orientation(int px, int py, int qx, int qy, int rx, int ry)
{
  // See https://www.geeksforgeeks.org/orientation-3-ordered-points/
  // for details of below formula.
  int val = (qy - py) * (rx - qx) - (qx - px) * (ry - qy);

  if (val == 0) return 0;  // colinear

  return (val > 0) ? 1 : 2; // clock or counterclock wise
}

// The main function that returns true if line segment 'p1q1' and 'p2q2' intersect.
int doIntersection(int p1x, int p1y, int q1x, int q1y, int p2x, int p2y, int q2x, int q2y)
{
  // Find the four orientations needed for general and special cases
  int o1 = orientation(p1x, p1y, q1x, q1y, p2x, p2y);
  int o2 = orientation(p1x, p1y, q1x, q1y, q2x, q2y);
  int o3 = orientation(p2x, p2y, q2x, q2y, p1x, p1y);
  int o4 = orientation(p2x, p2y, q2x, q2y, q1x, q1y);

  // General case
  if (o1 != o2 && o3 != o4)
    return 1; // true

  // Special Cases
  // p1, q1 and p2 are colinear and p2 lies on segment p1q1
  if (o1 == 0 && onSegment(p1x, p1y, p2x, p2y, q1x, q1y)) return 1; // true;

  // p1, q1 and q2 are colinear and q2 lies on segment p1q1
  if (o2 == 0 && onSegment(p1x, p1y, q2x, q2y, q1x, q1y)) return 1; // true;

  // p2, q2 and p1 are colinear and p1 lies on segment p2q2
  if (o3 == 0 && onSegment(p2x, p2y, p1x, p1y, q2x, q2y)) return 1; //true;

  // p2, q2 and q1 are colinear and q1 lies on segment p2q2
  if (o4 == 0 && onSegment(p2x, p2y, q1x, q1y, q2x, q2y)) return 1; // true;

  return 0; //false; Doesn't fall in any of the above cases
}

void generateMap(int matrix[8][8]) {
  int i;
  // store data about previously placed ships
  int sx[3] = { -1, -1, -1};  // x start
  int sy[3] = { -1, -1, -1};  // y start
  int d[3] = { -1, -1, -1};   // direction of ship
  int ex[3] = { -1, -1, -1};  // x end
  int ey[3] = { -1, -1, -1};  // y end

  // generate 3 ships of length 2, 3, 4
  for (i = 0; i < 3; i++) {
    int l = i + 2; // minimum length is 2
    int ok = 0;
    int xEnd, yEnd;
    while (ok == 0) {
      //get random direction : 0=up/1=down/2=left/3=right
      int dir = random(0, 4);
      // get random starting point: x & y
      int xStart = random(0, 8);
      int yStart = random(0, 8);

      if (sx[i] == xStart && sy[i] == yStart) {
        continue;
      }

      int lx = 0;
      int ly = 0;

      if (dir == 0) {
        // UP
        lx = -1;
      } else if (dir == 1) {
        // DOWN
        lx = 1;
      } else if (dir == 2) {
        // LEFT
        ly = -1;
      } else if (dir == 3) {
        // RIGHT
        ly = 1;
      }

      xEnd = xStart + lx * l;
      yEnd = yStart + ly * l;

      if (xEnd < 0 || yEnd < 0 || xEnd > 7 || yEnd > 7)
        continue;

      int j;
      int okInt = 0;
      for (j = 0; j < 3 && okInt == 0; j++) {
        if (d[j] == -1 || j == i) {
          continue;
        }
        int xs2 = sx[j];
        int ys2 = sy[j];
        int xe2 = xs2;
        int ye2 = ys2;
        if (d[j] == 0) {
          //up
          xe2 = xs2 - (j + 2);
        } else if (d[j] == 1) {
          // down
          xe2 = xs2 + (j + 2);
        } else if (d[j] == 2) {
          // left
          ye2 = ys2 - (j + 2);
        } else if (d[j] == 3) {
          //right
          ye2 = ys2 + (j + 2);
        }

        okInt = doIntersection(xStart, yStart, xEnd, yEnd, xs2, ys2, xe2, ye2);
      }
      if (okInt == 0) {
        // current ship doesn't intersec with previously placed ships
        ok = 1;
        // store data about the new ship
        sx[i] = xStart;
        sy[i] = yStart;
        ex[i] = xEnd;
        ey[i] = yEnd;
        d[i] = dir;
      }
    }
  }
  // mapping of the ships on the matrix
  for (i = 0; i < 3; i++ ) {
    if (ex[i] < sx[i] ) {
      int aux = ex[i];
      ex[i] = sx[i];
      sx[i] = aux;
    }
    if (ey[i] < sy[i]) {
      int aux = ey[i];
      ey[i] = sy[i];
      sy[i] = aux;
    }

    int l = i + 2;
    // writing "1"'s on  matrix
    int r, c;

    if (sx[i] == ex[i]) {
      for (c = sy[i]; c < ey[i]; c++) {
        matrix[sx[i]][c] = 1;
      }
    } else {
      for (r = sx[i]; r < ex[i]; r++) {
        matrix[r][sy[i]] = 1;
      }
    }
  }
}

// backup maps
void saveMaps() {
  for (int i = 0; i < 8; i++) {
    for (int j = 0; j < 8; j++) {
      saveA[i][j] = A[i][j];
      saveB[i][j] = B[i][j];
    }
  }
}

void generateMaps () {

  generateMap(A);

  generateMap(B);

  saveMaps();

  //print maps on serial monitor
  Serial.println("HARTA A");
  for (int i = 0; i < 8; i++) {
    for (int j = 0; j < 8; j++)
      Serial.print(A[i][j]);
    Serial.println();
  }

  Serial.println("\nHARTA B");
  for (int i = 0; i < 8; i++) {
    for (int j = 0; j < 8; j++)
      Serial.print(B[i][j]);
    Serial.println();
  }
}

void displayMapOnLED(int matrix[8][8]) {
  byte board[8] = {B00000000, B00000000, B00000000, B00000000, B00000000, B00000000, B00000000, B00000000};

  for (int i = 0; i < 8; i++) {
    for (int j = 0; j < 8; j++) {
      if (matrix[i][j] == 1) {
        int d = 7 - j;
        board[i] |= 1 << d;
      }
    }
  }
  for (int i = 0; i < 8; i++) {
    lc.setRow(0, i, board[i]);
  }
}

// player 0 attacks on map B, player 1 attacks on map A
int checkHit(int player, int x, int y) {
  // translate from the position selected by the player on the LED matrix to
  // the position mapping on the integer matrix with the actual map:
  // y is changing (the indexing is from right to left on the LED matrix)
  y = 7 - y;
  if (player == 0) {
    
    if (B[x][y] == 1) { // hit
      // modifiy initial matrix
      B[x][y] = 0;
      return 1;
    }
    else {

  tone(buzzer, 1000); // Send 1KHz sound signal...
  delay(300);        // ...for 1 sec
  noTone(buzzer);     // Stop sound...
  delay(300);        // ...for 1sec
  return 0; 
  }
  } else {
   
    if (A[x][y] == 1) {
      A[x][y] = 0;
      return 1;
    }
    else{
  tone(buzzer, 1000); // Send 1KHz sound signal...
  delay(300);        // ...for 1 sec
  noTone(buzzer);     // Stop sound...
  delay(300);        // ...for 1sec
  return 0; 
  }
  }

  return 0;
}

int checkWin(int player) {
  // if the matrix is full of 0's, then current player wins
  if (player == 0) {
    for (int i = 0; i < 8; i++) {
      for (int j = 0; j < 8; j++) {
        if (B[i][j] == 1) {
          // no win yet
          return 0;
        }
      }
    }
    // win
    return 1;
  } else {
    // same thing, but for player 1 on map A
    for (int i = 0; i < 8; i++) {
      for (int j = 0; j < 8; j++) {
        if (A[i][j] == 1) {
          // no win yet
          return 0;
        }
      }
    }
    // win

    return 1;
  }
}

// function to display the score on the LCD
void displayScore() {
  lcd.setCursor(0, 1);
  lcd.print("SCORE: ");
  lcd.print(scoreP0);
  lcd.print("-");
  lcd.print(scoreP1);
}

void loop() {

  // while game isn't won
  if (WON == -1) {
    char customKey = customKeypad.getKey();
    byte LEDboard[8] = {B00000000, B00000000, B00000000, B00000000, B00000000, B00000000, B00000000, B00000000};
    for (int i = 0; i < 8; i++) {
      for (int j = 0; j < 8; j++) {
        if (PLAYER == 0) {
          if (hitA[i][j] == 1) {
            int d = 7 - j;
            LEDboard[i] |= 1 << d;
          }
          if (missA[xPos][7 - yPos] == 1) {
            lcd.setCursor(12, 1);
            lcd.print("M");
          }
        }
        if (PLAYER == 1) {
          if (hitB[i][j] == 1) {
            int d = 7 - j;
            LEDboard[i] |= 1 << d;
          }
          if (missB[xPos][7 - yPos] == 1) {
            lcd.setCursor(12, 1);
            lcd.print("M");
          }
        }
      }
    }
    if (customKey) {
      // Clear LCD display and print characters
      lcd.clear();
      lcd.setCursor(0, 0);
      lcd.print("PLAYER #");
      lcd.print(PLAYER);
      lcd.print(" ");
      displayScore();

      // get current position as chosen by player through the keypad
      byte LEDboard[8] = {B00000000, B00000000, B00000000, B00000000, B00000000, B00000000, B00000000, B00000000};
      if (PLAYER == 0) {
        myServo.write(180);
        for (int i = 0; i < 8; i++) {
          for (int j = 0; j < 8; j++) {
            if (hitA[i][j] == 1) {
              int d = 7 - j;
              LEDboard[i] |= 1 << d;
            }
          }
        }
      }
      if (PLAYER == 1) {
        myServo.write(0);
        for (int i = 0; i < 8; i++) {
          for (int j = 0; j < 8; j++) {
            if (hitB[i][j] == 1) {
              int d = 7 - j;
              LEDboard[i] |= 1 << d;
            }
          }
        }
      }
      if (customKey == '2') { // UP
        lcd.setCursor(10, 0);
        lcd.print("UP");
        displayScore();
        if (xPos > 0)
          xPos--;
      } else if (customKey == '8') { // DOWN
        lcd.setCursor(10, 0);
        lcd.print("DOWN");
        displayScore();
        if (xPos < 7)
          xPos++;
      } else if (customKey == '4') { // LEFT
        lcd.setCursor(10, 0);
        lcd.print("LEFT");
        displayScore();
        if (yPos < 7)
          yPos++; // the indexing on the LED matrix is from right to left
      } else if (customKey == '6') { // RIGHT
        lcd.setCursor(10, 0);
        lcd.print("RIGHT");
        displayScore();
        if (yPos > 0)
          yPos--;
      } else if (customKey == '5') { // ATTACK
        // check if it was a hit/miss
        lcd.setCursor(10, 0);
        lcd.print("ATTACK");
        displayScore();
        delay(1000);
      
        
        

        lcd.clear();
        lcd.setCursor(0, 0);
        lcd.print("PLAYER #");
        lcd.print(PLAYER);
        // HIT or MISS
        if (checkHit(PLAYER, xPos, yPos) == 1) {
          // HIT
          lcd.print(" HIT");
          delay(1000);
          if (PLAYER == 0) {
            scoreP0++;
            yPos = 7 - yPos;
            A[xPos][yPos] = 0;
            hitA[xPos][yPos] = 1;
          }
          if (PLAYER == 1) {
            scoreP1++;
            yPos = 7 - yPos;
            B[xPos][yPos] = 1;
            hitB[xPos][yPos] = 1;
          }
          lcd.clear();
          lcd.setCursor(0, 0);
          lcd.print("PLAYER #");
          lcd.print(PLAYER);

          delay(1000);
          // WIN?
          if (checkWin(PLAYER) == 1) {
            lcd.clear();
            lcd.setCursor(0, 0);
            lcd.print("PLAYER #");
            lcd.print(PLAYER);
            lcd.print(" WON");
            winSound();
            WON = PLAYER;

            delay(1000);
          }
        } else {
          lcd.print(" MISS");
          //switch to next player
          if (PLAYER == 0) {
            PLAYER = 1;
            missA[xPos][7 - yPos] = 1;
          }

          else {
            PLAYER = 0;
            missB[xPos][7 - yPos] = 1;
          }
          delay(1000);
          lcd.clear();
          lcd.setCursor(0, 0);
          lcd.print("NEXT PLAYER #");
          lcd.print(PLAYER);
          delay(1000);
        }
      }
      if (customKey == '5') {
        displayScore();
        // it was an attack - no matter if it was a hit or a miss
        // reposition current player to start - upper left on the LED matrix
        // LEDboard[xPos]= B00111111;

        if (PLAYER == 0) {
          for (int i = 0; i < 8; i++) {
            for (int j = 0; j < 8; j++) {
              if (hitA[i][j] == 1) {
                int d = 7 - j;
                LEDboard[i] |= 1 << d;
              }
            }
          }
        }
        if (PLAYER == 1) {
          for (int i = 0; i < 8; i++) {
            for (int j = 0; j < 8; j++) {
              if (hitB[i][j] == 1) {
                int d = 7 - j;
                LEDboard[i] |= 1 << d;
              }
            }
          }
        }


        xPos = 0;
        yPos = 7;
        delay(500);

        LEDboard[xPos] |= 1 << yPos;
        for (int i = 0; i < 8; i++) {
          lc.setRow(0, i, LEDboard[i]);
        }
      }
      else if (customKey == '1') { // display map A
        lcd.clear();
        lcd.setCursor(0, 0);
          lcd.print("HARTA #0");
        displayMapOnLED(A);
        delay(2000);
      }
      else if (customKey == '3') { // display map b
        lcd.clear();
        lcd.setCursor(0, 0);
          lcd.print("HARTA #1");
        displayMapOnLED(B);
        delay(2000);
      }
       else if (customKey == '9') { // cheat
        lcd.clear();
        lcd.setCursor(0, 0);
        lcd.print("CHEAT MODE");
        cheat();
       }
    }
    LEDboard[xPos] |= 1 << yPos;
    for (int i = 0; i < 8; i++) {
      lc.setRow(0, i, LEDboard[i]);
    }
  } else {
    if (WON == 0) {
      // player 0 won -> display map B (defeated)
      displayMapOnLED(saveB);
    } else {
      // player 1 won -> display map A (defeated)
      displayMapOnLED(saveA);
    }
    
  }
}